'raytrace.inc based on C++ code
'by Jacco Bikker - jacco@bik5.com - www.bik5.com

$INCLUDE "raytypes.inc"

'scene
defint nPRIM
dim s1(4) as PRIM, s0 as PRIM

$INCLUDE "rayscene.inc"

dim E as ENGINE
'ray trace
dim light as PRIM
defint Depth=1
defsng Dist
dim pi3 as v3, L as v3, N as v3
'render frame
dim o1 as v3
dim o as v3, dir as v3, acc as v3
dim r0 as RAY  'ray from camera
defsng dot,diff,u
defint aMaterial,aPlane,aRay,aSphere

declare function GetSpecular as single
declare sub InitRender
declare function SphereIntersect as long
declare function PlaneIntersect as long
declare sub RayTrace
declare sub Render

'enter with aMaterial
function GetSpecular
defsng r
memcpy @r,iadd(aMaterial,12),4
result=1.0-r
end function

sub InitRender
E.Width=SCRWIDTH: E.Height=SCRHEIGHT
E.WX1=-4: E.WX2=4: E.WY1=3: E.WY2=-3
E.SY=3
E.DX=(E.WX2-E.WX1)/E.Width
E.DY=(E.WY2-E.WY1)/E.Height
E.SY=E.SY + E.DY
o.y=1: o.z=-4  'camera
o1=o  'for keyboard input
end sub

'enter with aSphere, aRay defined
function SphereIntersect
defsng b,det,i1,i2
dim v as v3
use ray as RAY with aRay
use sphere as SPHERE with aSphere
v.x=ray.pos.x-sphere.C.x
v.y=ray.pos.y-sphere.C.y
v.z=ray.pos.z-sphere.C.z
b = -(v.x*ray.dir.x + v.y*ray.dir.y + v.z*ray.dir.z)  '-DOT(v,dir)
'det = b*b - (v.x*v.x + v.y*v.y + v.z*v.z) + sphere.SqRadius
det = b*b + sphere.SqRadius - v.x*v.x - v.y*v.y - v.z*v.z)
defint res=MISS
if det>zero then
  det=sqr(det): i2=b+det
  if i2>zero then
    i1=b-det
    if i1<zero then
      if i2<Dist then Dist=i2: res=INPRIM
    else
      if i1<Dist then Dist=i1: res=HIT
    end if
  end if
end if
result=res
end use sphere
end use ray
end function

'enter with aPlane, aRay defined
function PlaneIntersect
defsng d0,d1
use ray as RAY with aRay
use plane as PLANE with aPlane
d0 = plane.N.x*ray.dir.x + plane.N.y*ray.dir.y + plane.N.z*ray.dir.z
result=MISS
if d0 then
  d1 = -(plane.N.x*ray.pos.x + plane.N.y*ray.pos.y + plane.N.z*ray.pos.z + plane.D)/d0
  if d1>zero and d1<Dist then Dist=d1: result=HIT
end if
end use plane
end use ray
end function

'Naive ray tracing: Intersects the ray with every primitive
'in the scene to determine the closest intersection.
'enter with aRay; Depth, RIndex 
sub RayTrace
defint hit,lite,s,prim
'if Depth>TRACEDEPTH then result=zero: exit function
'trace primary ray
Dist=1000000.0: prim=-1 
'find nearest intersection
for s=0 to nPRIM-one
s0=s1(s)  'get primitive
if s0.Type then
  aSphere=s0.O
  hit=SphereIntersect
else
  aPlane=s0.O
  hit=PlaneIntersect
end if
'prim = closest primative
if hit>zero then prim=s
next s
'no hit, terminate ray
if prim<zero then exit sub 

'handle intersection
s0=s1(prim)  'get hit primitive
if s0.Light then
  acc.x=1: acc.y=1: acc.z=1: exit sub 'light, stop tracing
end if

'determine color at point of intersection (pi3)
pi3.x = r0.pos.x + r0.dir.x*Dist
pi3.y = r0.pos.y + r0.dir.y*Dist
pi3.z = r0.pos.z + r0.dir.z*Dist
'trace lights
for lite = 0 to nPRIM-one
  if lite=prim then goto next_lite
  light=s1(lite)
  if light.Light then
    'calculate diffuse shading
    use sphere as SPHERE with light.O
    L.x=sphere.C.x-pi3.x: L.y=sphere.C.y-pi3.y: L.z=sphere.C.z-pi3.z
    'NORMALIZE(L)
    u=sqr(L.x*L.x + L.y*L.y + L.z*L.z)
    if u=zero then return
    L.x=L.x/u: L.y=L.y/u: L.z=L.z/u
    'GetNormal(prim,pi3)  'N gets normal v3
    memcpy @N,s0.O,12  'copy plane .N item to N
    if s0.Type then  'true if sphere
      asm mov eax,s0.O: asm mov eax,[eax+20]  'RRadius
      asm mov u,eax
      N.x=(pi3.x-N.x)*u
      N.y=(pi3.y-N.y)*u
      N.z=(pi3.z-N.z)*u
    end if
    dot=L.x*N.x + L.y*N.y + L.z*N.z
    if dot>zero then
      diff=dot*s0.M.Diff
      'add diffuse component to ray color
      acc.x=acc.x + diff*s0.M.c0*light.M.c0
      acc.y=acc.y + diff*s0.M.c1*light.M.c1
      acc.z=acc.z + diff*s0.M.c2*light.M.c2
    end if  
    end use sphere
  end if
next_lite:
next lite
end sub

'Fires rays in the scene one scanline at a time
'from left to right
sub Render
defint c
defsng u,d0
E.SY=E.WY1
for bmp_y=0 to E.Height-one
E.SX=E.WX1
'render pixels for current line
for bmp_x=0 to E.Width-one
 acc.x=zero: acc.y=zero: acc.z=zero 'use memset here
 dir.x=E.SX-o.x: dir.y=E.SY-o.y: dir.z=zero-o.z  'optimize this
 'Normalize(@dir)
 u=sqr(dir.x*dir.x + dir.y*dir.y + dir.z*dir.z)
 dir.x=dir.x/u: dir.y=dir.y/u: dir.z=dir.z/u
 r0.pos.x=o.x: r0.pos.y=o.y: r0.pos.z=o.z
 r0.dir.x=dir.x: r0.dir.y=dir.y: r0.dir.z=dir.z
 aRay=@r0
 RayTrace  '(@r0,@acc,1,1.0,@dist)
 $DEFINE gamma 400  '255
 defint Red=acc.x*gamma 
 defint Green=acc.y*gamma 
 defint Blue=acc.z*gamma 
 if Red>255 then Red=255
 if Green>255 then Green=255
 if Blue>255 then Blue=255
 rgb_c=rgb(Red,Green,Blue) 
 if rgb_c then SetPixel  'use bmp_x and bmp_y to put pixel rgb_c in bmp
 INC(E.SX,E.DX)
next bmp_x
INC(E.SY,E.DY): doevents
next bmp_y
end sub
